from .edge import Align, Edge
from .filein import FileIn, BatchIn, FileTest
from .fourier import Fourier
from .preedge import PreEdge
from .postedge import PostEdge
from .phase import Phase
from .plot import Plot
from .background import Background
from .othercmds import Cut, Smooth, Rebin, Normalize
from .glitch import Deglitch, MultiEdge
from .save import Save
from .pca import PCA
from .average import Average

from ._handler import CommandHandler

commands: dict[str, CommandHandler] = {
    "filein": FileIn(),
    "batchin": BatchIn(),
    "align": Align(),
    "edge": Edge(),
    "preedge": PreEdge(),
    "postedge": PostEdge(),
    "fourier": Fourier(),
    "phase": Phase(),
    "plot": Plot(),
    "background": Background(),
    "cut": Cut(),
    "smooth": Smooth(),
    "rebin": Rebin(),
    "deglitch": Deglitch(),
    "multiedge": MultiEdge(),
    "save": Save(),
    "test": FileTest(),
    "pca": PCA(),
    "average": Average(),
    "normalize": Normalize(),
}



import argparse

from typing import NamedTuple
from ._exceptions import InputFileParsingException
from ._numberunit import NUMBER_UNIT

class Token(NamedTuple):
    value: str
    line: int
    pos: int


class CommandParser(argparse.ArgumentParser):
    def __init__(
        self,
        prog: str | None = None,
        usage: str | None = None,
        description: str | None = None,
        epilog: str | None = None,
        **kwargs
    ) -> None:
        super().__init__(
            prog, usage, description, epilog, **kwargs, exit_on_error=False
        )

    def parse(
        self, tokens: list[Token], _namespace: None | argparse.Namespace = None
    ) -> argparse.Namespace:
        args = [token.value for token in tokens]
        # Dirty workaround: if the value is of the form NUMBER_UNIT then replace
        # the "-" with "!THISISANEGATIVE!" temporarily. parse_nu can handle this prefix.
        args = [a if not a.startswith("-") else 
                (a if NUMBER_UNIT.match(a) is None else "!THISISANEGATIVE!"+a.removeprefix("-")) for a in args]

        try:
            namespace = self.parse_args(args, _namespace)
        except argparse.ArgumentError as E:
            raise InputFileParsingException(E)

        return namespace  # type: ignore



import colorlog
import logging
import os
import sys

from argparse import ArgumentParser
from pathlib import Path
from tomllib import loads as toml_loads

from . import __version__

from .commands._context import Paths, Context, Options
from .parser import parse_version, parse_directives, parse_commands

from matplotlib import pyplot as plt

logging.getLogger("matplotlib").setLevel(logging.WARNING)
logging.getLogger("PIL").setLevel(logging.WARNING)


def main():
    parser = ArgumentParser(
        "estrapy", description="Analyze XAS data files from an instruction file."
    )

    parser.add_argument("inputfile", nargs="?", help="The instruction file.")
    parser.add_argument(
        "outputdir", nargs="?", help="The output directory. If empty, creates one."
    )
    parser.add_argument(
        "--verbose", "-v", action="store_true", help="Makes the output more complete."
    )
    parser.add_argument(
        "--legacy", "-l", help="Runs the legacy version of the program."
    )
    parser.add_argument(
        "--version", "-V", action="version", version=f"EstraPy {__version__}"
    )
    parser.add_argument(
        "--no-color", action="store_true", help="Removes color from the log."
    )
    parser.add_argument(
        "--no-update",
        action="store_true",
        help="Does not update the input file if an old version is detected.",
    )
    parser.add_argument(
        "--no-cache", action="store_true", help="Does not read or store from the cache."
    )
    parser.add_argument("--vars", nargs="+", help="Sets runtime variables, accessed as %1% or ${1} in the input file.")

    args = parser.parse_args()

    # --- Set context ----------------------------------------------------------

    currentdir = Path(os.getcwd())

    userdir = Path.home() / ".estrapy"
    userdir.mkdir(exist_ok=True)

    configfile = userdir / "config.toml"
    configfile.touch(exist_ok=True)

    match args.inputfile, args.outputdir:
        case ("-", _o):
            # Interactive
            print("Interactive mode is not implemented.")
            exit(-1)

        case (None, None):
            # Search the current directory for input files
            print("Automatic discovery mode is not implemented.")
            exit(-1)

        case (_i, None):
            inputfile = Path(_i)
            outputdir = inputfile.parent / inputfile.name.removesuffix(inputfile.suffix)

        case (None, _o):
            print("Impossible input combination.")
            exit(-1)

        case (_i, _o):
            inputfile = Path(_i)
            outputdir = Path(_o)

    if not inputfile.exists():
        print("The specified file does not exist.")
        exit(-1)

    outputdir.mkdir(exist_ok=True)
    logfile = outputdir / (inputfile.name.removesuffix(inputfile.suffix) + ".log")

    paths = Paths(
        inputfile,
        currentdir,
        inputfile.parent,
        outputdir,
        userdir,
        logfile,
        configfile,
    )
    config = toml_loads(configfile.read_text("utf-8"))

    inputfiledata = inputfile.read_text("utf-8")
    directives = parse_directives(inputfiledata)

    if args.vars is not None:        
        for i,var in enumerate(args.vars, 1):
            directives.vars[str(i)] = var

    options = Options(
        False,
        args.verbose,
        parse_version(inputfiledata),
        not args.no_color,
        not args.no_update,
        not args.no_cache,
        {},
    )
    context = Context(paths, config, options, directives)

    # --- Logging --------------------------------------------------------------

    rootlogger = colorlog.getLogger()
    rootlogger.setLevel(logging.DEBUG)

    # Rename all levels to align in the log file
    [
        logging.addLevelName(v, n)
        for n, v in (("DBG", 10), ("INF", 20), ("WRN", 30), ("ERR", 40), ("FTL", 50))
    ]
    log_colors = {
        "DBG": "light_black",
        "INF": "white",
        "WRN": "yellow",
        "ERR": "red",
        "FTL": "bold_red",
    }

    fmt = logging.Formatter("[%(levelname)s] <%(name)s> %(message)s")
    flh = logging.FileHandler(logfile, encoding="utf-8", mode="w" if directives.clear else "a")
    flh.setLevel(logging.DEBUG)
    flh.setFormatter(fmt)
    rootlogger.addHandler(flh)

    if options.logcolor:
        cfm = colorlog.ColoredFormatter(
            "%(log_color)s[%(levelname)s] <%(name)s> %(message)s", log_colors=log_colors
        )
        sth = colorlog.StreamHandler(sys.stdout)
        sth.setLevel(logging.DEBUG if options.verbose else logging.INFO)
        sth.setFormatter(cfm)
    else:
        sth = logging.StreamHandler(sys.stdout)
        sth.setLevel(logging.DEBUG if options.verbose else logging.INFO)
        sth.setFormatter(fmt)
    rootlogger.addHandler(sth)

    log = logging.getLogger("estrapy")

    # --------------------------------------------------------------------------

    log.info(f" EstraPy version {__version__} ".center(64, "-"))
    log.debug(f"Input file is {paths.inputfile}")
    log.debug(f"Working directory is {paths.workdir}")
    log.debug(f"Output directory is {paths.outputdir}")

    # --- Parse input file -----------------------------------------------------

    commands = parse_commands(inputfiledata, context)

    for linenumber, executor, commandargs in commands:
        try:
            res = executor.execute(commandargs, context)
        except Exception as E:
            log.critical(f"Error at line {linenumber}: {E}")
            break
        context.commands.append((commandargs, res))
    pass

    # At the end of the commands, check if there are any figures that are not shown.
    if any([not fig.shown for fig in context.figures.figureruntimes.values()]):
        plt.show()




from __future__ import annotations

import numpy as np
import numpy.typing as npt

from enum import Enum
from logging import getLogger
from typing import NamedTuple

from ._context import Context, Column, FourierType, Domain
from ._handler import CommandHandler, Token, CommandResult

from ._parser import CommandParser


def extract_phase(
    x: npt.NDArray[np.floating], y: npt.NDArray[np.complexfloating]
) -> tuple[npt.NDArray[np.complexfloating], np.poly1d]:
    """Extracts the linear behavior of the phase.
    Returns:
      - the corrected complex signal
      - the removed linear component as a np.poly1d
    """
    # Numpy unwrap the phase and linear fit to remove most of the windings
    p0 = np.unwrap(np.angle(y))
    p_p0 = np.poly1d(np.polyfit(x, p0, 1))
    linear_phase = p_p0(x)
    y1 = y * (np.cos(linear_phase) - 1.0j * np.sin(linear_phase))

    # Phase and phase derivative evaluation for better unwrapping

    # p1 = np.unwrap(np.angle(y1))
    # dp1 = np.diff(p1, prepend=[0])
    # _s = (np.cumsum(dp1 > np.pi/2) - np.cumsum(dp1 < - np.pi/2)) * np.pi
    # s = np.cos(_s) + 1.0j*np.sin(_s)
    # y2 = y1 * s.conj()

    # Rotate the phase so that the center of mass is real
    c = (_t := y1.mean()) / np.abs(_t)

    y3 = y1 * c.conj()

    return y3, p_p0 - np.angle(c)


class PhaseAction(Enum):
    CORRECT = "corr"
    ALIGN = "align"


class Args_Phase(NamedTuple):
    action: PhaseAction


class Phase(CommandHandler):
    @staticmethod
    def parse(tokens: list[Token], context: Context) -> Args_Phase:
        parser = CommandParser(
            "phase", description="Alters the phase of the fourier transform."
        )
        subparsers = parser.add_subparsers(dest="action")

        correct = subparsers.add_parser("correct")
        align = subparsers.add_parser("align")

        args = parser.parse(tokens)

        match args.action:
            case "correct":
                action = PhaseAction.CORRECT
            case "align":
                action = PhaseAction.ALIGN
            case _:
                raise RuntimeError("Invalid phase transformation.")

        return Args_Phase(action)

    @staticmethod
    def execute(args: Args_Phase, context: Context) -> CommandResult:
        log = getLogger("phase")

        match args.action:
            case PhaseAction.CORRECT:
                for data in context.data:
                    r = data.get_col_("R", domain=Domain.FOURIER)
                    f = data.get_col_("f", domain=Domain.FOURIER)
                    corrected, phase = extract_phase(r, f) # type: ignore

                    P = phase(r)

                    data.mod_col("f", corrected)
                    data.add_col("pcorr", P, Column(None, None, FourierType.PHASE), Domain.FOURIER)
                    data.meta.run["phasecorrect"] = phase
                    log.debug(f"{data.meta.name}: Removed linear phase contribution.")

            case PhaseAction.ALIGN:
                for dataidx, data in enumerate(context.data):
                    # Take the first data, then align everything to it
                    if dataidx == 0:
                        firstdata = data.get_xy("R", "f")
                        data.mod_col("f", firstdata.to_numpy())
                        continue
                    
                    f = data.get_xy("R", "f")
                    dot = np.vdot(firstdata, f)
                    c = dot / np.abs(dot)
                    data.mod_col("f", f * c.conj())
                    log.debug(f"{data.meta.name}: Aligned all spectra to minimize distance to first.")


        return CommandResult(True)

    @staticmethod
    def undo(args: NamedTuple, context: Context) -> CommandResult:
        # TODO:
        raise NotImplementedError



from typing import NamedTuple
from logging import getLogger

from . import __version__
from .commands import commands
from .commands._context import Directives, Context
from .commands._handler import CommandHandler
from .commands._parser import InputFileParsingException

VERSION = tuple(int(i) for i in __version__.split("."))


def parse_version(input: str) -> tuple[int, ...]:
    # At most the version line will be 25 characters long. This is in order to
    # reduce the amount of unnecessary string splitting.
    firstline = input[:25].splitlines()[0]
    if not firstline.startswith("#"):
        raise InputFileParsingException(
            "The first line of the input file must declare the version."
        )

    versionstr = (
        firstline.removeprefix("#")
        .strip()
        .removeprefix("version")
        .removeprefix(":")
        .strip()
    )

    return tuple(int(d) for d in versionstr.split("."))


def parse_directives(input: str) -> Directives:
    lines = input.splitlines()
    # Directives are prefixed by %, and follow the version. Only empty lines can
    # be between directives or version, and all directives must be at the top of
    # the file.

    clear = False
    noplot = False
    vars = {}

    for line in lines[1:]:
        # Skip empty lines
        if not line.strip():
            continue
        # Stop directive parsing after the first command line
        if not line.startswith("%"):
            break
        
        directive = line.removeprefix("%").strip()
        if directive == "clear":
            clear = True
        if directive == "noplot":
            noplot = True

        if directive.startswith("define "):
            vname, vval = directive.removeprefix("define ").split(" ", maxsplit=1)
            vars[vname] = vval

    return Directives(clear, noplot, vars)


def parse_commands(
    input: str, context: Context
) -> list[tuple[int, CommandHandler, NamedTuple]]:
    log = getLogger("parser")

    # Check file version
    if context.options.version[:2] != VERSION[:2]:
        # TODO: is not valueerror
        raise ValueError(
            f"Estrapy version ({VERSION}) does not match file version {context.options.version}"
        )
    
    replaced = input
    for varname, varval in context.directives.vars.items():
        replaced = replaced.replace(f"%{varname}%", varval)
        replaced = replaced.replace(f"${{{varname}}}", varval)

    lines = [*enumerate(replaced.splitlines(), 1)]

    parsedcommands: list[tuple[int, CommandHandler, NamedTuple]] = []

    lineid = 0
    while lineid + 1 < len(lines):
        lineid += 1
        index, line = lines[lineid]
        if not line.strip():
            continue
        if line.startswith("#"):
            continue
        if line.startswith("%"):
            continue

        if line.startswith(" "):
            raise InputFileParsingException("Command cannot start with spaces.")

        cmd = line.split(maxsplit=1)[0]

        if cmd == "exit":
            return parsedcommands
        
        if cmd not in commands:
            raise InputFileParsingException(f"Unrecognized command: {cmd}")

        command = commands[cmd]

        # Check if successive lines start with spaces and are non empty, and add
        # them to the current command line.
        step = 1
        while (
            (lineid + step < len(lines))
            and (lines[lineid + step][1].startswith(" "))
            and lines[lineid + step][1].strip()
        ):
            step += 1

        tokens = command.tokenize([(i, lines[i][1]) for i in range(lineid, lineid + step)])
        # Removes the first token, corresponding to the command name.
        try:
            commandargs = command.parse(tokens[1:], context)
        except Exception as E:
            log.critical(f"Syntax error: line {index} {str(E)}")
            exit(-1)

        parsedcommands.append((index, command, commandargs))

        # -1 is due to the +1 at the beginning of the loop
        lineid += step - 1

    return parsedcommands



from __future__ import annotations

import numpy as np

from typing import NamedTuple
from logging import getLogger

from ._context import Context, Column, DataColType, AxisType, Domain, range_to_index
from ._format import pol
from ._handler import CommandHandler, Token, CommandResult
from ._numberunit import parse_range, NumberUnitRange, actualize_range

from ._parser import CommandParser


class Args_PreEdge(NamedTuple):
    bounds: NumberUnitRange
    degree: int

class PreEdge(CommandHandler):
    @staticmethod
    def parse(tokens: list[Token], context: Context) -> Args_PreEdge:
        parser = CommandParser(
            "preedge", description="Removes the background preedge contribution."
        )
        parser.add_argument("range", nargs=2, help="The polynomial fit range.")

        groupd = parser.add_mutually_exclusive_group()
        groupd.add_argument(
            "--constant",
            "-C",
            dest="degree",
            action="store_const",
            const=0,
            help="Models the preedge as a constant polynomial.",
        )
        groupd.add_argument(
            "--linear",
            "-l",
            dest="degree",
            action="store_const",
            const=1,
            help="Models the preedge as a linear polynomial.",
        )
        groupd.add_argument(
            "--quadratic",
            "-q",
            dest="degree",
            action="store_const",
            const=2,
            help="Models the preedge as a quadratic polynomial.",
        )
        groupd.add_argument(
            "--cubic",
            "-c",
            dest="degree",
            action="store_const",
            const=3,
            help="Models the preedge as a cubic polynomial.",
        )
        groupd.add_argument(
            "--polynomial",
            "-p",
            dest="degree",
            type=int,
            help="Models the preedge as a polynomial trend, and removes it from the data.",
        )

        args = parser.parse(tokens)

        range = parse_range(*args.range)
        if range.domain != Domain.RECIPROCAL:
            raise ValueError("Invalid fit domain: the preedge can only be calculated in energy or wavevector.")

        return Args_PreEdge(range, args.degree)

    @staticmethod
    def execute(args: Args_PreEdge, context: Context) -> CommandResult:
        log = getLogger("preedge")

        domain = args.bounds.domain or Domain.RECIPROCAL
        if domain != Domain.RECIPROCAL:
            raise RuntimeError("Cannot fit preedge to a different domain.")
        
        _axes = [data.get_col_(data.datums[domain].default_axis) for data in context.data] # type: ignore
        range = actualize_range(args.bounds, _axes, "eV")

        for data in context.data:
            if "preedge" in data.meta.run:
                raise RuntimeError(f"Preedge was already calculated for {data.meta.name}")
            idx = range_to_index(data, range)
            log.debug(f"{data.meta.name}: Fitting preedge of order {args.degree} in the region {range.lower.value:0.3f}{range.lower.unit} ~ {range.upper.value:0.3f}{range.upper.unit}") # type: ignore

            X,Y = data.get_col_("e"), data.get_col_("a")
            x, y = X[idx], Y[idx]
            poly = np.poly1d(np.polyfit(x, y, args.degree)) # type: ignore
            P = poly(X)

            log.debug(f"{data.meta.name}: preedge = {pol(poly.coef)}")

            data.meta.run["preedge"] = poly, AxisType.RELENERGY
            data.add_col("pre", P, Column(None, None, DataColType.PREEDGE), Domain.RECIPROCAL)
            data.mod_col("a", Y - P)
            pass

        return CommandResult(True)

    @staticmethod
    def undo(args: NamedTuple, context: Context) -> CommandResult:
        # TODO:
        raise NotImplementedError



from __future__ import annotations

import numpy as np
import pandas as pd

from typing import NamedTuple
from logging import getLogger

from ._context import Context, Column, DataColType, AxisType, Domain, range_to_index
from ._format import pol
from ._handler import CommandHandler, Token, CommandResult
from ._numberunit import parse_range, NumberUnitRange
from ._numberunit import actualize_range

from ._parser import CommandParser


class Args_PCA(NamedTuple):
    ncomponents: int

class PCA(CommandHandler):
    @staticmethod
    def parse(tokens: list[Token], context: Context) -> Args_PCA:
        parser = CommandParser(
            "pca", description="Performs PCA decomposition on the data."
        )
        parser.add_argument("--numcomponents", "-n", type=int, default=10)

        args = parser.parse(tokens)

        return Args_PCA(args.numcomponents)

    @staticmethod
    def execute(args: Args_PCA, context: Context) -> CommandResult:
        log = getLogger("pca")
        from matplotlib import pyplot as plt

        _X = pd.DataFrame({i:data.get_xy("E", "mu") for i,data in enumerate(context.data)}).T
        X = _X.loc[:,_X.isna().sum() == 0]
        m = X.mean()
        Xm = X - m

        ev,eV = np.linalg.eigh(Xm.cov())
        ev = pd.Series(ev[::-1], index=[f"PC{i+1}" for i in range(Xm.shape[1])])
        eV = pd.DataFrame(eV[:,::-1], columns=ev.index, index = X.columns)

        pass
        return CommandResult(True)

    @staticmethod
    def undo(args: NamedTuple, context: Context) -> CommandResult:
        # TODO:
        raise NotImplementedError



from __future__ import annotations

import numpy as np
import numpy.typing as npt
import pandas as pd

from logging import getLogger
from typing import NamedTuple
from dataclasses import dataclass
from statsmodels.nonparametric.smoothers_lowess import lowess
from scipy.stats import norm
from scipy.interpolate import interp1d

from ._numberunit import NumberUnitRange, parse_range, actualize_range, parse_nu

from ._context import Context, Column, Domain, range_to_index, DataColType
from ._handler import CommandHandler, Token, CommandResult

from ._parser import CommandParser


@dataclass(slots=True, frozen=True)
class Finder:...

@dataclass(slots=True, frozen=True)
class Finder_Force(Finder):...

@dataclass(slots=True, frozen=True)
class Finder_Variance(Finder):
    width: int
    pvalue: float

@dataclass(slots=True, frozen=True)
class Finder_Smooth(Finder):
    width: int
    fraction: float
    pvalue: float

@dataclass(slots=True, frozen=True)
class Finder_Polynomial(Finder):
    width: int
    degree: int
    pvalue: float

@dataclass(slots=True, frozen=True)
class Finder_Even(Finder):
    pvalue: float


@dataclass(slots=True, frozen=True)
class Method:
    noise:bool

@dataclass(slots=True, frozen=True)
class Method_Remove(Method):
    ...

@dataclass(slots=True, frozen=True)
class Method_Base(Method):
    ...

@dataclass(slots=True, frozen=True)
class Method_Smooth(Method):
    fraction: float



class Args_Deglitch(NamedTuple):
    bounds: NumberUnitRange
    column: str
    finder: Finder
    method: Method


class Deglitch(CommandHandler):
    @staticmethod
    def parse(tokens: list[Token], context: Context) -> Args_Deglitch:
        parser = CommandParser(
            "deglitch", description="Removes and replaces glitched signals."
        )
        parser.add_argument("range", nargs=2)
        parser.add_argument("--column", "-c", default="I0")
        # parser.add_argument("finder", choices=["force", "variance", "smooth", "polynomial", "even"], default="polynomial")
        # parser.add_argument("--window", "-w", type=int)
        # parser.add_argument("--degree", "-d", type=int)
        # parser.add_argument("--column", "-c", default="I0")
        # parser.add_argument("--pvalue", "-p", type=float, default=3)
        subparsers = parser.add_subparsers(dest="finder")
        force = subparsers.add_parser("force")

        variance = subparsers.add_parser("variance")
        variance.add_argument("--width", "-w", type=int, default=10)
        variance.add_argument("--pvalue", "-p", type=float, default=0.001)
        
        smooth = subparsers.add_parser("smooth")
        smooth.add_argument("--fraction", "-f", type=float, default=0.3)
        smooth.add_argument("--width", "-w", type=int, default=10)
        smooth.add_argument("--pvalue", "-p", type=float, default=0.001)
        
        polynomial = subparsers.add_parser("polynomial")
        polynomial.add_argument("--degree", "-d", type=int, default=1)
        polynomial.add_argument("--width", "-w", type=int, default=10)
        polynomial.add_argument("--pvalue", "-p", type=float, default=0.001)
        
        even = subparsers.add_parser("even")
        even.add_argument("--pvalue", "-p", type=float, default=0.001)

        for sp in subparsers._name_parser_map.values():
            subparsers = sp.add_subparsers(dest="method")
            remove = subparsers.add_parser("remove")
            remove.add_argument("--noise", "-n", action="store_true")

            base = subparsers.add_parser("base")
            base.add_argument("--noise", "-n", action="store_true")

            smooth = subparsers.add_parser("smooth")
            smooth.add_argument("--noise", "-n", action="store_true")
            smooth.add_argument("--fraction", "-f", type=float, default=0.05)

        args = parser.parse(tokens)

        _range = parse_range(*args.range)
        if _range.domain == None:
            _range = NumberUnitRange(_range.lower, _range.upper, _range.inter, Domain.RECIPROCAL)
        if _range.domain != Domain.RECIPROCAL:
            raise ValueError("Cannot deglitch on non-reciprocal domain.")

        match args.finder:
            case "force":
                finder = Finder_Force()
            case "variance":
                finder = Finder_Variance(args.width, args.pvalue)
            case "smooth":
                finder = Finder_Smooth(args.width, args.fraction, args.pvalue)
            case "polynomial":
                finder = Finder_Polynomial(args.width, args.degree, args.pvalue)
            case "even":
                finder = Finder_Even(args.pvalue)
            case f:
                raise ValueError(f"Invalid finder: {f}")
        
        match args.method:
            case "remove":
                method = Method_Remove(args.noise)
            case "base":
                if args.column != "a":
                    raise ValueError("Method base can only be used if column is a.")
                method = Method_Base(args.noise)
            case "smooth":
                method = Method_Smooth(args.noise, args.fraction)
            case f:
                raise ValueError(f"Invalid method: {f}")

        return Args_Deglitch(
            _range,
            args.column,
            finder,
            method
        )

    @staticmethod
    def execute(args: Args_Deglitch, context: Context) -> CommandResult:
        log = getLogger("deglitch")

        domain = args.bounds.domain or Domain.RECIPROCAL
        if domain != Domain.RECIPROCAL:
            raise RuntimeError("Cannot fit preedge to a different domain.")
        
        _axes = [data.get_col_(data.datums[domain].default_axis) for data in context.data] # type: ignore
        range = actualize_range(args.bounds, _axes, "eV")

        for data in context.data:
            idx = range_to_index(data, range)
            X, I = data.get_col_("E"), data.get_col_(args.column)
            x,i = X[idx], I[idx]
            g = np.zeros_like(idx)
            b:npt.NDArray[np.floating] | None = None
            std = None
            
            match args.finder:
                case Finder_Force():
                    g[:] = idx
                case Finder_Variance(width, pvalue):
                    # Estimate the std as median of stds
                    std = pd.Series(i,x).rolling(len(X) // width, center=True).std().median()
                    u = norm.ppf(1-pvalue/2) * std

                    g[idx] = np.abs(i) > u
                case Finder_Smooth(width, fraction, pvalue):
                    # Assume I0 is represented by a low frequency trend
                    #    I0 = I0' + eps
                    # Model I0' as a severely smoothed I0
                    b = lowess(i, x, fraction, it=0, return_sorted=False)
                    y = i-b

                    # Estimate the std as median of stds
                    std = pd.Series(y,x).rolling(len(X) // width, center=True).std().median()
                    u = norm.ppf(1-pvalue/2) * std
                    g[idx] = np.abs(y) > u
                    pass
                case Finder_Polynomial(width, degree, pvalue):
                    # Assume I0 is represented by:
                    #    I0 = P_n + eps
                    # Model the I0 trend as a n-polynomial
                    poly = np.poly1d(np.polyfit(x,i, degree))
                    b = poly(x)
                    y = i-b
                    # Estimate the std as median of stds
                    std = pd.Series(y,x).rolling(len(X) // width, center=True).std().median()
                    u = norm.ppf(1-pvalue/2) * std

                    g[idx] = np.abs(y) > u
                    pass
                case Finder_Even(pvalue):
                    # Determine the point difference between consecutive points
                    idxeven = np.indices(x.shape)[0] % 2 == 0
                    xe,ie = x[idxeven], i[idxeven]
                    xo,io = x[~idxeven], i[~idxeven]
                    ieo = interp1d(xe, ie, "cubic", bounds_error=False)(xo)
                    ioe = interp1d(xo, io, "cubic", bounds_error=False)(xe)
                    ii = np.zeros_like(i)
                    ii[idxeven] = ioe
                    ii[~idxeven] = ieo

                    y = i - ii
                    b = ii
                    # Estimate the std as median of stds
                    std = pd.Series(y,x).rolling(len(X) // 10, center=True).std().median()
                    u = norm.ppf(1-pvalue/2) * std
                    g[idx] = np.abs(y) > u
            
            if (_sum:=sum(g)) > 1:
                log.debug(f"{data.meta.name}: Found glitch in the region {range.lower.value}{range.lower.unit} ~ {range.upper.value}{range.upper.unit} of width {_sum} points.")

                match (args.method, b, std):
                    case Method_Remove(), _, _:
                        data.datums[domain].df = data.datums[domain].df[~idx]
                        log.debug(f"{data.meta.name}: Removed glitched region.")
                    case (Method_Base(noise), b, std):
                        if b is None or std is None:
                            raise RuntimeError("Method base requires estimation of the baseline and standard deviation.")
                        if args.column != "a":
                            raise RuntimeError("Method base can only be used if the column is a")
                        newcol = I.copy()
                        if noise: n = np.random.normal(0, std, g.sum())
                        else: n = np.zeros(g.sum())
                        
                        newcol[g] = b[g[idx]]+n
                        data.mod_col("a", newcol)
                        log.debug(f"{data.meta.name}: Removed glitch, replaced with baseline data.")
                    case (Method_Smooth(noise, fraction), _, _):
                        x = data.get_col_("E", domain=domain)
                        y = data.get_col_("a", domain=domain)
                        s = lowess(y[~g],x[~g], it=0, frac=fraction, return_sorted=False, xvals=x)
                        std = (y-s)[idx].std()
                        if noise: n = np.random.normal(0, std, g.sum())
                        else: n = np.zeros(g.sum())
                        
                        yn = y.copy()
                        yn[g] = s[g]+n
                        log.debug(f"{data.meta.name}: Removed glitch, replaced with smoothed data.")
            

        return CommandResult(True)

    @staticmethod
    def undo(args: NamedTuple, context: Context) -> CommandResult:
        # TODO:
        raise NotImplementedError

class ArcTangent(NamedTuple):
    a: float
    b: float
    c: float
    def calc(self, x:npt.NDArray[np.floating]) -> npt.NDArray[np.floating]:
        s = np.tan(0.4 * np.pi)
        # c is defined to be the shift from b to be 10% - 90%
        return self.a * (np.atan(s * (x-self.b)/abs(self.c)) / np.pi + 0.5)
        # return self.a * (np.atan((x-self.b)/self.c) / np.pi + 0.5) # not reparametrized

class HyperTangent(NamedTuple):
    a: float
    b: float
    c: float
    def calc(self, x:npt.NDArray[np.floating]) -> npt.NDArray[np.floating]:
        s = np.atanh(0.9)
        return self.a * (np.tanh(s * (x-self.b)/abs(self.c)) + 1) * 0.5


class ErrorFunction(NamedTuple):
    a: float
    b: float
    c: float
    def calc(self, x:npt.NDArray[np.floating]) -> npt.NDArray[np.floating]:
        from scipy.special import erf, erfinv
        s = erfinv(0.8)
        return self.a * (1 + erf(s * (x-self.b)/abs(self.c))) * 0.5

class Exponential(NamedTuple):
    a: float
    b: float
    c: float
    def calc(self, x:npt.NDArray[np.floating]) -> npt.NDArray[np.floating]:
        y = np.zeros_like(x)
        s = np.log(0.1)
        i = x > self.b

        if self.c > 0:
            y[i] = self.a * (1 - np.exp(s * (x[i]-self.b)/self.c))
        elif self.c < 0:
            y[i] = self.a
            y[~i] = self.a * (np.exp(s * (x[~i]-self.b)/self.c))
        else:
            y[i] = 1

        return y


class Args_MultiEdge(NamedTuple):
    axis: str
    raw: bool
    method: ArcTangent|ErrorFunction|Exponential|HyperTangent

class MultiEdge(CommandHandler):
    @staticmethod
    def parse(tokens: list[Token], context: Context) -> Args_MultiEdge:
        parser = CommandParser(
            "multiedge", description="Removes and replaces multiple edges."
        )
        parser.add_argument("--energy", "-E", dest="axis", const="E", action="store_const")
        parser.add_argument("--relenergy", "-e", dest="axis", const="e", action="store_const")
        parser.add_argument("--kvector", "-k", dest="axis", const="k", action="store_const")
        parser.add_argument("--raw", "-r", action="store_true")

        subparsers = parser.add_subparsers(dest="mode")

        arctan = subparsers.add_parser("atan")
        errfun = subparsers.add_parser("erf")
        tanhyp = subparsers.add_parser("tanh")
        expfun = subparsers.add_parser("exp")

        for _parser in [arctan, errfun, tanhyp, expfun]:
            _parser.add_argument("b")
            _parser.add_argument("a")
            _parser.add_argument("c")

        args = parser.parse(tokens)

        a = parse_nu(args.a).value
        b = parse_nu(args.b).value
        c = parse_nu(args.c).value

        match args.mode:
            case "atan": mode = ArcTangent(a, b, c)
            case "erf": mode = ErrorFunction(a, b, c)
            case "exp": mode = Exponential(a, b, c)
            case "tanh": mode = HyperTangent(a, b, c)

        axis = args.axis if args.axis is not None else "E"
        
        return Args_MultiEdge(axis, args.raw, mode)

    @staticmethod
    def execute(args: Args_MultiEdge, context: Context) -> CommandResult:
        log = getLogger("multiedge")

        for data in context.data:
            X = data.get_col_(args.axis, domain=Domain.RECIPROCAL)

            E = args.method.calc(X)
            # TODO: idea --iplot interactive plot:
            # Si apre un plot e un terminale interattivo.
            # L'utente può cambiare i parametri della curva partendo da quelli
            # dati nel file, per ottimizzare il residuo, che viene plottato ad
            # ogni cambio. Alla fine, stampa la linea del comando finale.

            data.datums[Domain.RECIPROCAL].add_col(E, Column(None, None, DataColType.MULTIEDGE), "mult", unique=False)
            
            if args.raw:
                a = data.get_col_("a")
                data.datums[Domain.RECIPROCAL].mod_col("a", a-E)

                if "mu" in data.datums[Domain.RECIPROCAL].cols and "x" in data.datums[Domain.RECIPROCAL].cols:
                    J0 = data.meta.get("J0")
                    if not isinstance(J0, float):
                        raise RuntimeError("AAhkfbwefdjisjndba")
                    mu = data.get_col_("mu")
                    chi = data.get_col_("x")
                    data.datums[Domain.RECIPROCAL].mod_col("mu", mu-E/J0)
                    data.datums[Domain.RECIPROCAL].mod_col("x", chi-E/J0)
            else:
                # if raw is not given, assume norm was called, so J0, mu and chi exist
                J0 = data.meta.get("J0")
                if not isinstance(J0, float):
                    raise RuntimeError("AAhkfbwefdjisjndba")
                a = data.get_col_("a")
                mu = data.get_col_("mu")
                chi = data.get_col_("x")

                data.datums[Domain.RECIPROCAL].mod_col("a", a-E*J0)
                data.datums[Domain.RECIPROCAL].mod_col("mu", mu-E)
                data.datums[Domain.RECIPROCAL].mod_col("x", chi-E)

            match args.method:
                case ArcTangent(a, b, c): msg = f"atan({b}, {a}, {c})"
                case ErrorFunction(a, b, c): msg = f"erf({b}, {a}, {c})"
                case Exponential(a, b, c): msg = f"exp({b}, {a}, {c})"
                case HyperTangent(a, b, c): msg = f"tanh({b}, {a}, {c})"
            log.debug(f"{data.meta.name}: Removed multiple excitation contribution as {msg}")
        return CommandResult(True)

    @staticmethod
    def undo(args: NamedTuple, context: Context) -> CommandResult:
        # TODO:
        raise NotImplementedError




import re
import numpy as np
import numpy.typing as npt

from typing import NamedTuple, Sequence
from enum import Enum

SI_PREFIXES = ("kMGTPEZYRQ", "munpfazyrq")
SI_MULTIPLIER:dict[str,int] = {
    pname:(1000 ** pval) * (1 if sign == 0 else -1)
    for sign,prefixes in enumerate(SI_PREFIXES)
    for pval,pname in enumerate(prefixes,1)
}

def SI_multiplier(prefix: str | None) -> int:
    if prefix is None: return 1
    return SI_MULTIPLIER.get(prefix, 1)

_ELEMENT_EDGE = r"(\w+)\.(\w+)(?:([+-]?\d+(?:\.\d+)?)(?:([kMGTPEZYRQmunpfazyrq])?(eV))?)?"
ELEMENT_EDGE = re.compile(_ELEMENT_EDGE)

_NUMBER_UNIT = r"^(([+-]?)?\d+(?:\.\d+)?)(?:([kMGTPEZYRQmunpfazyrq])?([A-Za-z]+))?$"
NUMBER_UNIT = re.compile(_NUMBER_UNIT)


class Domain(Enum):
    RECIPROCAL = "reciprocal"
    FOURIER = "fourier"
    PCA = "pca"


class NumberUnit(NamedTuple):
    value: float
    sign: int
    unit: str | None

class Bound(Enum):
    INTERNAL = 0
    EXTERNAL = 1

class NumberUnitRange(NamedTuple):
    lower: NumberUnit | Bound
    upper: NumberUnit | Bound
    inter: NumberUnit | int | None
    domain: Domain | None

def parse_nu(val: str) -> NumberUnit:
    if val.startswith("!THISISANEGATIVE!"):
        return parse_nu("-"+val.removeprefix("!THISISANEGATIVE!"))
    m = NUMBER_UNIT.match(val)
    if m is None:
        raise ValueError(f'Could not parse "{val}".')
    
    _value, _sign, _mult, _unit = m.groups()
    # TODO: TMP:
    # if _unit is None: raise ValueError("Temporary error")
    value = float(_value) * SI_multiplier(_mult)
    sign = ["-", "", "+"].index(_sign) -1

    return NumberUnit(value, sign, _unit)

def get_unit_domain(v:NumberUnit) -> Domain | None:
    if v.unit is None: return None
    return {
        "eV": Domain.RECIPROCAL,
        "k": Domain.RECIPROCAL,
        "q": Domain.RECIPROCAL,
        "A": Domain.FOURIER
    }.get(v.unit)

def _parse_lowbound(val:str) -> tuple[NumberUnit | Bound, Domain | None]:
    match val:
        case "..":
            return NumberUnit(-np.inf, 0, None), None
        case ".:":
            return Bound.INTERNAL, None
        case ":.":
            return Bound.EXTERNAL, None
        case _:
            lower = parse_nu(val)
            return lower, get_unit_domain(lower)
def _parse_higbound(val:str) -> tuple[NumberUnit | Bound, Domain | None]:
    match val:
        case "..":
            return NumberUnit(np.inf, 0, None), None
        case ".:":
            return Bound.EXTERNAL, None
        case ":.":
            return Bound.INTERNAL, None
        case _:
            upper = parse_nu(val)
            return upper, get_unit_domain(upper)

def parse_range(*vals:str, number:int|None=None) -> NumberUnitRange:
    match vals:
        case _lower, _upper:
            lower, ldomain = _parse_lowbound(_lower)
            upper, udomain = _parse_higbound(_upper)

            match lower, upper:
                case NumberUnit(lv,_,u), NumberUnit(uv,us,None) if lv != -np.inf and uv == np.inf:
                    upper = NumberUnit(np.inf, us, u)
                case NumberUnit(lv,ls,None), NumberUnit(uv,_,u) if lv == -np.inf and uv != np.inf:
                    lower = NumberUnit(-np.inf, ls, u)

            match ldomain, udomain:
                case None, None:
                    return NumberUnitRange(lower, upper, number, None)
                case [domain, None] | [None, domain]:
                    return NumberUnitRange(lower, upper, number, domain)
                case ld,ud if ld == ud:
                    return NumberUnitRange(lower, upper, number, ld)
                case _: raise ValueError("Unit domain mismatch.")

        case _lower, _upper, _interval:
            inter = parse_nu(_interval)
            lower, ldomain = _parse_lowbound(_lower)
            upper, udomain = _parse_higbound(_upper)

            if inter.unit == "n":
                # If unit is n, interval is the number of points
                if number is not None: raise ValueError("Two distinct number of points specifications.")

                match ldomain, udomain:
                    case None, None:
                        return NumberUnitRange(lower, upper, int(inter.value), None)
                    case [domain, None] | [None, domain]:
                        return NumberUnitRange(lower, upper, int(inter.value), domain)
                    case ld,ud if ld == ud:
                        return NumberUnitRange(lower, upper, int(inter.value), ld)
                    case _: raise ValueError("Unit domain mismatch.")
            else:
                idomain = get_unit_domain(inter)

                match ldomain, udomain, idomain:
                    case None, None, None:
                        return NumberUnitRange(lower, upper, inter, None)
                    case [domain, None, None] | [None, domain, None] | [None, None, domain]:
                        return NumberUnitRange(lower, upper, inter, domain)
                    case [a,b,None] | [a,None,b] | [None,a,b] if a==b:
                        return NumberUnitRange(lower, upper, inter, a)
                    case ld, ud, id if (ld == ud == id):
                        return NumberUnitRange(lower, upper, inter, id)
                    case _: raise ValueError("Unit domain mismatch.")
    
    raise ValueError("Invalid range.")

def actualize_lower(lowerbound: NumberUnit|Bound, axes:Sequence[npt.NDArray], unit:str|None) -> NumberUnit:
    match lowerbound:
        case Bound.INTERNAL: return NumberUnit(max([np.min(ax) for ax in axes]), 0, unit)
        case Bound.EXTERNAL: return NumberUnit(min([np.min(ax) for ax in axes]), 0, unit)
        case NumberUnit(unit=None): return NumberUnit(lowerbound.value, lowerbound.sign, unit)
        case NumberUnit(): return lowerbound
    raise RuntimeError("Invalid bound.")

def actualize_upper(upperbound: NumberUnit|Bound, axes:Sequence[npt.NDArray], unit:str|None) -> NumberUnit:
    match upperbound:
        case Bound.INTERNAL: return NumberUnit(min([np.max(ax) for ax in axes]), 0, unit)
        case Bound.EXTERNAL: return NumberUnit(max([np.max(ax) for ax in axes]), 0, unit)
        case NumberUnit(unit=None): return NumberUnit(upperbound.value, upperbound.sign, unit)
        case NumberUnit(): return upperbound
    raise RuntimeError("Invalid bound.")

def actualize_range(range:NumberUnitRange, axes:Sequence[npt.NDArray], unit:str|None) -> NumberUnitRange:
    lower, upper = actualize_lower(range.lower, axes, unit), actualize_upper(range.upper, axes, unit)
    return NumberUnitRange(lower, upper, range.inter, range.domain)



from __future__ import annotations

import numpy as np
import numpy.typing as npt

# from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from scipy.interpolate import interp1d
from matplotlib import pyplot as plt



from enum import Enum
from typing import NamedTuple, Any
from statsmodels.nonparametric.smoothers_lowess import lowess
from scipy.optimize import minimize_scalar, OptimizeResult, root_scalar, RootResults
from logging import getLogger
from dataclasses import dataclass

from ._context import Context, Column, AxisType, FigureSettings, FigureRuntime
from ._handler import CommandHandler, Token, CommandResult
from ._misc import E_to_sk, parse_edgeenergy, nderivative
from ._numberunit import NumberUnit, parse_nu, Domain
from ._parser import CommandParser


class Operation(Enum):
    CUT = "c"

    POLYFIT = "p"
    SMOOTH = "s"
    DERIVATIVE = "d"
    INTERPOLATE = "i"

    MAXIMUM = "M"
    MINIMUM = "m"
    ZERO = "Z"
    HALFHEIGHT = "H"
    AVERAGEINT = "A"
    SET = "S"


FINAL_METHODS = [
    Operation.MAXIMUM,
    Operation.MINIMUM,
    Operation.ZERO,
    Operation.HALFHEIGHT,
    Operation.AVERAGEINT,
    Operation.SET,
]

@dataclass(slots=True, frozen=True)
class Method:...

@dataclass(slots=True, frozen=True)
class Finder(Method):
    operations: list[tuple[Operation, int|None]]
    E0s: float | None
    dE0: float

@dataclass(slots=True, frozen=True)
class Across(Method):
    resolution: int
    dE0: float
    # corrshift: int


class Args_Align(NamedTuple):
    method: Method
    # method: list[tuple[Operation, int | None]]
    E0: float

    wplot: bool


class Args_Edge(NamedTuple):
    method: Method

    wplot: bool

def _get_norm(x:npt.NDArray, shiftcenter:bool=False) -> tuple[float, float]:
    m,M = np.min(x), np.max(x)
    if shiftcenter: return m, M-m
    else:           return 0, max(abs(M), abs(m))

def norm(x:npt.NDArray, center:float, factor:float) -> npt.NDArray:
    return (x-center) / factor

def find_E0_with_method(
    method: list[tuple[Operation, int | None]],
    _bounds: tuple[float, float],
    _y: npt.NDArray[np.floating],
    _x: npt.NDArray[np.floating],
    *, ax:None|Axes=None
) -> float:
    E0shift = (_bounds[1] + _bounds[0])/2

    h: list[tuple[str, Any]] = [("data", (_y, _x - E0shift))]
    _current_bound: tuple[float, float] = np.min(_x) - E0shift, np.max(_x) - E0shift # type: ignore
    _deriv_order = 0
    _norm_factor:dict[int, tuple[float, float]] = {}

    bounds = _bounds[0] - E0shift, _bounds[1] - E0shift

    if ax is not None:
        ax.axvspan(*bounds, alpha=0.1)
        

    for action, arg in method:
        dtype, data = h[-1]

        match action, dtype:
            case [Operation.CUT, "data"]:
                arg = arg or 0
                y, x = data  # type: ignore
                y: npt.NDArray[np.floating]
                x: npt.NDArray[np.floating]
                assert arg is not None
                d = arg * (bounds[1] - bounds[0]) / 2
                a, b = bounds[0] - d, bounds[1] + d

                idx = (x >= a) & (x <= b)
                pad = np.pad(idx, (1, 1))
                idx = idx | pad[2:] | pad[:-2]
                h.append(("data", (y[idx], x[idx])))
                _current_bound = a,b

                if ax:
                    if _deriv_order not in _norm_factor:
                        _norm_factor[_deriv_order] = _get_norm(y[idx], _deriv_order == 0)
                    
                    ax.plot(x[idx], norm(y[idx], *_norm_factor[_deriv_order]), label=f"c{arg}")

            case [Operation.POLYFIT, "data"]:
                if arg is None:
                    continue
                y, x = data  # type: ignore
                y: npt.NDArray[np.floating]
                x: npt.NDArray[np.floating]
                poly = np.poly1d(np.polyfit(x, y, arg))
                h.append(("poly", poly))

                if ax:
                    _pX = np.linspace(*_current_bound, 1001)
                    ax.plot(_pX, norm(poly(_pX), *_norm_factor[_deriv_order]), label=f"p{arg}")

            case [Operation.SMOOTH, "data"]:
                if arg is None:
                    continue
                y, x = data  # type: ignore
                y: npt.NDArray[np.floating]
                x: npt.NDArray[np.floating]
                y = lowess(y, x, arg / len(x), it=0, return_sorted=False)
                h.append(("data", (y, x)))

                if ax:
                    ax.plot(x, norm(y, *_norm_factor[_deriv_order]), label=f"s{arg}")

            case [Operation.DERIVATIVE, "data"]:
                if arg is None:
                    continue
                y, x = data  # type: ignore
                y: npt.NDArray[np.floating]
                x: npt.NDArray[np.floating]
                y = nderivative(y, arg, x) # type: ignore
                _deriv_order += arg
                h.append(("data", (y, x)))

                if ax:
                    if _deriv_order not in _norm_factor:
                        _norm_factor[_deriv_order] = _get_norm(y, _deriv_order == 0)
                    ax.plot(x, norm(y, *_norm_factor[_deriv_order]), label=f"d{arg}")

            case [Operation.DERIVATIVE, "poly"]:
                if arg is None:
                    continue
                poly = data
                poly: np.poly1d
                poly = poly.deriv(arg)
                _deriv_order += arg

                h.append(("poly", poly))

                if ax:
                    _pX = np.linspace(*_current_bound, 1001)
                    if _deriv_order not in _norm_factor:
                        _norm_factor[_deriv_order] = _get_norm(poly(_pX), _deriv_order == 0)

                    ax.plot(_pX, norm(poly(_pX), *_norm_factor[_deriv_order]), label=f"p{arg}")

            case [Operation.INTERPOLATE, "data"]:
                if arg is None:
                    continue
                y, x = data  # type: ignore
                y: npt.NDArray[np.floating]
                x: npt.NDArray[np.floating]
                interp = interp1d(x, y, arg, fill_value=np.nan) # type: ignore
                h.append(("interp", interp))

                if ax:
                    _pX = np.linspace(*_current_bound, 1001)
                    ax.plot(_pX, norm(interp(_pX), *_norm_factor[_deriv_order]), label=f"i{arg}")

            case [Operation.MAXIMUM, "data"]:
                y, x = data  # type: ignore
                y: npt.NDArray[np.floating]
                x: npt.NDArray[np.floating]

                M = x[y.argmax()]
                if ax: ax.axvline(M)
                return M + E0shift

            case [Operation.MINIMUM, "data"]:
                y, x = data  # type: ignore
                y: npt.NDArray[np.floating]
                x: npt.NDArray[np.floating]

                m = x[y.argmin()]
                if ax: ax.axvline(m)
                return m + E0shift

            case [Operation.MAXIMUM, "poly" | "interp"]:
                p = data
                res: OptimizeResult = minimize_scalar(-p, bounds=bounds)  # type: ignore
                if ax: ax.axvline(res.x)
                return res.x + E0shift

            case [Operation.MINIMUM, "poly" | "interp"]:
                p = data
                res: OptimizeResult = minimize_scalar(p, bounds=bounds)  # type: ignore
                if ax: ax.axvline(res.x)
                return res.x + E0shift

            case [Operation.ZERO, "data"]:
                y, x = data  # type: ignore
                y: npt.NDArray[np.floating]
                x: npt.NDArray[np.floating]
                Z = x[np.abs(y).argmin()]
                if ax:
                    ax.axvline(Z)
                    ax.axhline(0, linestyle=":")
                return Z + E0shift

            case [Operation.ZERO, "poly" | "interp"]:
                p = data
                try:
                    res: RootResults = root_scalar(p, bracket=bounds)  # type: ignore
                except ValueError:
                    # Let's try again with the bounded one, but ensuring
                    # that the bracket has different values and includes the
                    # middle value.
                    _b0, _b1 = bounds
                    for i in range(10):
                        if p(_b0) * p(_b1) <= 0: break
                        _b0, _b1 = (_b1-_b0)*0.1, _b1 - (_b1-_b0)*0.1
                    else:
                        raise ValueError("Cannot find root interval.")
                    res: RootResults = root_scalar(p, bracket=(_b0, _b1))  # type: ignore
                if ax:
                    ax.axvline(res.root)
                    ax.axhline(0, linestyle=":")
                return res.root + E0shift

            case [Operation.HALFHEIGHT, "data"]:
                raise NotImplementedError("This operation was not yet implemented.")
            case [Operation.AVERAGEINT, "data"]:
                raise NotImplementedError("This operation was not yet implemented.")
            case [Operation.SET, _]:
                return E0shift
            case [op, dt]:
                raise RuntimeError(f"Invalid operation: {op} on {dt}")
    
    raise ValueError("Method must terminate with a final method.")


def _parse_method_to_op(method: str) -> tuple[Operation, int | None]:
    if len(method) == 1:
        m, a = method, None
    elif len(method) == 2:
        m, a = method[0], int(method[1])
    else:
        raise ValueError(f'Invalid method specification: "{method}"')
    match m:
        case "c":
            return (Operation.CUT, a)
        case "p":
            return (Operation.POLYFIT, a)
        case "s":
            return (Operation.SMOOTH, a)
        case "d":
            return (Operation.DERIVATIVE, a)
        case "i":
            return (Operation.INTERPOLATE, a)

        case "M":
            return (Operation.MAXIMUM, a)
        case "m":
            return (Operation.MINIMUM, a)
        case "Z":
            return (Operation.ZERO, a)
        case "H":
            return (Operation.HALFHEIGHT, a)
        case "A":
            return (Operation.AVERAGEINT, a)
        case "S":
            return (Operation.SET, a)
        case _:
            raise ValueError(f'Invalid method specification: "{method}"')


def parse_method_to_ops(method: str) -> list[tuple[Operation, int | None]]:
    methods = [_parse_method_to_op(m) for m in method.split(".")]
    if any(m[0] in FINAL_METHODS for m in methods[:-1]):
        raise ValueError("Final method in non-final position.")
    if methods[-1][0] not in FINAL_METHODS:
        raise ValueError("Method must terminate with a final method.")

    return methods

_method_aliases = {
    "set": "S",
    "fitderivative": "c1.s5.d1.p3.M",
    "fitpolynomial": "c1.p3.d1.M",
    "fitmaximum": "c1.p3.M",
    "interpderivative": "c1.s5.d1.i3.M",
    "maximum": "c1.i3.M",
}


class Align(CommandHandler):
    @staticmethod
    def parse(tokens: list[Token], context: Context) -> Args_Align:
        parser = CommandParser("align", description="Aligns the reference data to a standard or a reference E0 value.")
        subparser = parser.add_subparsers(dest="method")
        at = subparser.add_parser("at")
        at.add_argument("operations", help="The series of operations to find the E0.")
        at.add_argument("--E0", "-E", type=parse_edgeenergy, required=True, help="Sets the E0 value for the edge.")
        at.add_argument("--dE0", "-d", help="Interval search width.", required=True)
        at.add_argument("--search","-s",type=parse_edgeenergy,help="Searches around this value. If not set, searches around E0")
        at.add_argument("--wplot", action="store_true", help="Plots the result of the edge search.")

        shift = subparser.add_parser("shift")
        shift.add_argument("--dE0", "-d", help="Interval search width.", required=True)
        shift.add_argument("--resolution", default=30, type=int, help="The resolution multiplier.")
        shift.add_argument("--wplot", action="store_true", help="Plots the result of the edge search.")


        # parser.add_argument("method", help="Uses the specified method to find the E0.")
        # parser.add_argument("--E0", "-E", type=parse_edgeenergy, required=True, help="Sets the E0 value for the edge.")
        # parser.add_argument("--dE0", "-d", type=parse_nu, help="Interval search width.")
        # parser.add_argument("--search","-s",type=parse_edgeenergy,help="Searches around this value. If not set, searches around E0")
        # parser.add_argument("--wplot", action="store_true", help="Plots the result of each edge search.")
        # parser.add_argument("--rplot", action="store_true", help="Plots the result of all edge searches together.")
        # parser.add_argument("--resolution", default=30, type=int)


        args = parser.parse(tokens)

        match args.method:
            case "at":
                E0 = args.E0
                search = args.search if args.search is not None else E0
                dE0 = parse_nu(args.dE0)
                if dE0.unit != "eV":
                    raise ValueError(f'Unit mismatch: unit "{dE0.unit}" is not compatible with eV.')
                dE0 = dE0.value

                if args.operations in _method_aliases:
                    ops = _method_aliases[args.operations]
                else: ops = args.operations
                
                method = Finder(parse_method_to_ops(ops), search, dE0)

            case "shift":
                dE0 = parse_nu(args.dE0)
                if dE0.unit != "eV":
                    raise ValueError(f'Unit mismatch: unit "{dE0.unit}" is not compatible with eV.')
                dE0 = dE0.value
                method = Across(args.resolution, dE0)
                E0 = None

            case _: raise ValueError("Unknown method")

        return Args_Align(
            method,
            E0,
            args.wplot,
        )

    @staticmethod
    def execute(args: Args_Align, context: Context) -> CommandResult:
        log = getLogger("align")

        match args.method:
            case Finder(ops, E0s, dE0) if E0s is not None:
                for data in context.data:
                    y = data.get_col_("ref")
                    x = data.get_col_("E")

                    if args.wplot:
                        fignum = context.figures.get_high_figurenum()
                        # TODO: remove FigureRuntime and FigureSettings
                        figure = FigureRuntime.new(FigureSettings(fignum, (1,1)))
                        context.figures.figureruntimes[fignum] = figure
                        ax = figure.axes[(1,1)].axis
                        ax.set_title(f"Reference E0 detection for {data.meta.name}")
                        ax.set_xlabel("Energy shift")
                    else: ax = None

                    try:
                        rE0 = find_E0_with_method(ops, (E0s - dE0, E0s + dE0), y, x, ax=ax)
                    except ValueError:
                        log.error(f"{data.meta.name}: Cannot find reference E0 value with the specified method.")
                        continue
                
                    # Check where the value falls. If it is outside the range, error
                    _l, _u = E0s - dE0, E0s + dE0
                    if (rE0 <= _l) or (rE0 >= _u):
                        log.error(f"{data.meta.name}: E0 value found outside the specified range ({rE0:.3f}eV).")
                        continue
                    elif (rE0 <= _l + 0.05*dE0) or (rE0 >= _u - 0.05*dE0):
                        log.warning(f"{data.meta.name}: E0 value found very close to the specified range ({rE0:.3f}eV)")
                    # If it is in the ±5% of the borders of the range, warning
                
                    if ax is not None:
                        ax.legend()
                        figure.show()

                    shift = rE0 - args.E0

                    log.info(
                        f"{data.meta.name}: Found reference E0 value at {rE0:.3f}eV (shift by {-shift:+.3f}eV)"
                    )

                    data.meta.refE0 = args.E0
                    data.mod_col("E", x - shift)

                    if data.meta.E0 is not None:
                        data.meta.E0 = data.meta.E0 - shift

            case Across(resolution, dE0):
                rXY = [data.get_xy_("E", "ref") for data in context.data]

                # Get common interpolation X axis
                m, M = max(np.min(rX) for rX,_ in rXY), min(np.max(rX) for rX,_ in rXY)
                dx = np.median(np.concat([np.diff(rX) for rX,_ in rXY]))
                newdx = dx/resolution
                cX = np.arange(m, M, newdx)

                # Interpolate the gradients onto the common grid
                drXY = [(rX, np.gradient(rY)) for rX, rY in rXY]
                dYs = [interp1d(rX, drY, "cubic")(cX) / np.std(drY) for rX, drY in drXY]

                # Calculate the point shift from the dE0
                corrshift = int(np.ceil(dE0 / newdx))
                corrs = [np.correlate(dY, dYs[0][corrshift:-corrshift], "valid") for dY in dYs]
                cx = np.linspace(-newdx * corrshift, newdx * corrshift, len(corrs[0]))

                shifts = np.array([np.poly1d(np.polyfit(cx, corr, 2)).deriv().roots[0] for corr in corrs])
                shifts = shifts - np.mean(shifts)

                for data, shift in zip(context.data, shifts):
                    data.mod_col("E", data.get_col_("E") - shift)
                    data.meta.refE0 = args.E0

                if args.wplot:
                    fignum = context.figures.get_high_figurenum()
                    fig = plt.figure(fignum)
                    ax11 = fig.add_subplot(2,2,1)
                    ax11.hist(shifts, bins=30)

                    ax12 = fig.add_subplot(2,2,2)
                    for c in corrs:
                        ax12.plot(cx, c)
                    
                    ax21 = fig.add_subplot(2,2,3)
                    ax21.plot(shifts)

                    ax21 = fig.add_subplot(2,2,4)
                    for rX, rY in rXY:
                        plt.plot(rX, rY, color="blue")
                    for rX, rY in rXY:
                        plt.plot(rX, rY, color="red", alpha=0.5, linewidth=1)
                    

                
        return CommandResult(True)

    @staticmethod
    def undo(args: NamedTuple, context: Context) -> CommandResult:
        # TODO:
        raise NotImplementedError


class Edge(CommandHandler):
    @staticmethod
    def parse(tokens: list[Token], context: Context) -> Args_Edge:
        parser = CommandParser("edgeenergy", description="Aligns the reference data to a standard or a reference E0 value.")
        parser.add_argument("operations", help="Uses the specified method to find the E0.")
        parser.add_argument("--E0", "-E", type=parse_edgeenergy, help="Searches around this value. If not set, searches around reference E0.")
        parser.add_argument("--dE0", "-d", type=parse_nu, help="Interval search width.")
        parser.add_argument("--wplot", action="store_true", help="Plots the result of each edge search.")

        args = parser.parse(tokens)

        E0s:float|None = args.E0

        if args.dE0 is not None:
            _dE0 = args.dE0
            _dE0: NumberUnit
            if _dE0.unit != "eV":
                raise ValueError(f'Unit mismatch: unit "{_dE0.unit}" is not compatible with eV.')
            dE0 = _dE0.value
        else:
            dE0 = 5.0

        match args.operations:
            case str(ops):
                if ops in _method_aliases:
                    ops = _method_aliases[ops]
                method = Finder(parse_method_to_ops(ops), E0s, dE0)

        return Args_Edge(
            method,
            args.wplot,
        )

    @staticmethod
    def execute(args: Args_Edge, context: Context) -> CommandResult:
        log = getLogger("edge")

        for data in context.data:
            if data.meta.E0 is not None:
                raise RuntimeError(
                    f"Reference preedge was already calculated for {data.meta.name}."
                )
            
            match args.method:
                case Finder(ops, E0s, dE0):
                    x = data.get_col_("E")
                    y = data.get_col_("a")

                    if E0s is None:
                        E0s = data.meta.refE0
                        if E0s is None:
                            raise RuntimeError("E0 search interval was not provided.")

                    if args.wplot:
                        fignum = context.figures.get_high_figurenum()
                        figure = FigureRuntime.new(FigureSettings(fignum, (1,1)))
                        context.figures.figureruntimes[fignum] = figure
                        ax = figure.axes[(1,1)].axis
                        ax.set_title(f"E0 detection for {data.meta.name}")
                        ax.set_xlabel("Energy shift")
                    else: ax = None

                    try:
                        E0 = find_E0_with_method(ops, (E0s - dE0, E0s + dE0), y, x, ax=ax) # type: ignore
                    except ValueError:
                        log.error(f"{data.meta.name}: Cannot find E0 value with the specified method.")
                        if ax is not None:
                            ax.legend()
                            figure.show()
                        continue

                    # Check where the value falls. If it is outside the range, error
                    _l, _u = E0s - dE0, E0s + dE0
                    if (E0 <= _l) or (E0 >= _u):
                        log.error(f"{data.meta.name}: E0 value found outside the specified range ({E0:.3f}eV).")
                        continue
                    elif (E0 <= _l + 0.05*dE0) or (E0 >= _u - 0.05*dE0):
                        log.warning(f"{data.meta.name}: E0 value found very close to the specified range ({E0:.3f}eV)")
                    # If it is in the ±5% of the borders of the range, warning
                
                    if ax is not None:
                        ax.legend()
                        figure.show()
                    
                    if data.meta.refE0:
                        relativeE0 = E0 - data.meta.refE0
                        log.info(
                            f"{data.meta.name}: Found E0 value at {E0:.3f}eV ({relativeE0:+.3f}eV from reference)"
                        )
                    else:
                        log.info(f"{data.meta.name}: Found E0 value at {E0:.3f}eV")

                    data.meta.E0 = E0
                case _: raise RuntimeError("Invalid state: #23578423")
            
            data.add_col("e", np.array(x - E0), Column("eV", True, AxisType.RELENERGY), Domain.RECIPROCAL)
            data.add_col("k", E_to_sk(x, E0), Column("k", None, AxisType.KVECTOR), Domain.RECIPROCAL)

        return CommandResult(True)

    @staticmethod
    def undo(args: NamedTuple, context: Context) -> CommandResult:
        # TODO:
        raise NotImplementedError



from __future__ import annotations

from typing import NamedTuple
from logging import getLogger

from ._context import Context, Domain, range_to_index, DataColType, Column
from ._handler import CommandHandler, Token, CommandResult
from ._numberunit import NumberUnit, parse_range, NumberUnitRange, actualize_range
from ._parser import CommandParser


class Shell(NamedTuple):
    feff_file: str
    

class Args_Fit(NamedTuple):
    bounds: NumberUnitRange


class Fit(CommandHandler):
    @staticmethod
    def parse(tokens: list[Token], context: Context) -> Args_Fit:
        parser = CommandParser("fit", description="Fits the experimental data to a number of shells, with given parameters.")
        parser.add_argument("range", nargs=2)
        parser.add_argument("--mode", "-m", choices=["k", "r"], default="k", help="Fitting mode: k-space or r-space. Default is k-space.")
        parser.add_argument("--dir", "-d", type=str, default="./feff", help="Directory where the FEFF files are located. Default is in a separate feff folder.")
        parser.add_argument("--param", action="append", nargs="+", help="Define a fitting parameter. Usage: --param name initial [min max step]. Can be used multiple times.")
        parser.add_argument("--shell", action="append", nargs="+", help="Define a fitting shell. Usage: --shell feff_file -param1 param1 -param2 param2 ... Can be used multiple times.")

        args = parser.parse(tokens)

        return Args_Fit()

    @staticmethod
    def execute(args: Args_Fit, context: Context) -> CommandResult:
        log = getLogger("fit")

        domain = args.bounds.domain or Domain.RECIPROCAL
        _axes = [data.get_col_(data.datums[domain].default_axis) for data in context.data] # type: ignore
        range = actualize_range(args.bounds, _axes, "eV")
        
        for data in context.data:
            idx = range_to_index(data, range)
            data.datums[domain].df = data.datums[domain].df.loc[idx, :]
                
            log.debug(f"{data.meta.name}: Cut data in the range {range.lower.value:0.3f}{range.lower.unit} ~ {range.upper.value:0.3f}{range.upper.unit}") # type: ignore

        return CommandResult(True)

    @staticmethod
    def undo(args: NamedTuple, context: Context) -> CommandResult:
        # TODO:
        raise NotImplementedError



from __future__ import annotations

import numpy as np

from logging import getLogger
from typing import NamedTuple, Iterable, TypeVar
from functools import reduce
from scipy.interpolate import interp1d

from ._context import Context, Data, DataStore, MetaData, SignalType, AxisType
from ._handler import CommandHandler, Token, CommandResult
from ._parser import CommandParser

_T = TypeVar("_T")

def eqor(s:Iterable[_T], *, default:_T = ...) -> _T:
    _it = iter(s)
    try: s0 = next(_it)
    except StopIteration:
        if default is ...: raise
        return default
    for e in _it:
        if e != s0:
            if default is ...: raise ValueError("All elements are not equal.")
            return default
    return s0

class Args_Average(NamedTuple):
    groupvars: list[str]
    axis: str | None

class Average(CommandHandler):
    @staticmethod
    def parse(tokens: list[Token], context: Context) -> Args_Average:
        parser = CommandParser("average", description="Averages data.")
        parser.add_argument("--by", "-b", nargs="+")
        parser.add_argument("--axis", "-a")

        args = parser.parse(tokens)

        return Args_Average(groupvars=args.by or [], axis=args.axis)

    @staticmethod
    def execute(args: Args_Average, context: Context) -> CommandResult:
        log = getLogger("average")

        bins:dict[tuple, list[Data]] = {}
        for outdata in context.data:
            var = tuple(outdata.meta.get(v, default=None) for v in args.groupvars)
            if var not in bins: bins[var] = []
            bins[var].append(outdata)
        
        newdata = DataStore()
        for var,datas in bins.items():

            # Define metadata for the new data
            ivars:set[tuple[str, str|int|float]] = reduce(set.intersection, [{*data.meta.vars.items()} for data in datas])
            avars:set[tuple[str, str|int|float]] = reduce(set.union, [{*data.meta.vars.items()} for data in datas])
            mvars:dict[str, list[int|float]] = {}
            for k,v in avars.difference(ivars):
                if k.startswith("."): continue
                if isinstance(v, str): continue
                if k not in mvars: mvars[k] = []
                mvars[k].append(v)
            vvars: dict[str, int|float] = {k:sum(vs)/len(vs) for k,vs in mvars.items()}
            nvars = dict(ivars) | vvars

            signaltype = eqor([data.meta.signaltype for data in datas], default = SignalType.OTHER)
            refsigtype = eqor([data.meta.refsigtype for data in datas], default = SignalType.OTHER)
            name = "_".join(str(v) for v in var)
            path = context.paths.workdir
            refE0 = eqor([data.meta.refE0 for data in datas], default = None)

            run = {"mean": datas}
            
            E0s = [data.meta.E0 for data in datas]
            if all(E is not None for E in E0s):
                E0 = sum(E0s) / len(E0s) # type: ignore
            else: E0 = None

            # Add the {.fn} variable as the "_".join of the variables
            nvars[".fn"] = name
            outmeta = MetaData(signaltype, refsigtype, name, path, nvars, run, refE0, E0)
            outdata = Data(outmeta)

            # TODO: Temo che al momento non sia possibile fare le cose bene.
            # Ciò richiederebbe di definire un modo per convertire da `E` a `e`
            # e `k` univoco, definendo a necessità `E0` per ciascuno spettro mediato.
            # Al momento faccio le cose "male", prendendo come nuovi assi gli 
            # assi del primo (entro i range degli altri)

            datums = set(d for _d in datas for d in _d.datums)
            for datum in datums:
                firstdata, *_ = [data.datums[datum] for data in datas]
                # Get the first data axes and add them to data
                for colname, colkind in firstdata.cols.items():
                    if isinstance(colkind.axis, AxisType):
                        outdata.add_col(colname, firstdata.df[colname], colkind, domain=datum)
                
                if args.axis is None:
                    daxisname, daxis = firstdata.default_axis, firstdata.get_default_axis()
                else:
                    daxisname, daxis = args.axis, firstdata.df[args.axis]

                for colname, colkind in firstdata.cols.items():
                    # Skip the columns that are already present
                    if colname in outdata.datums[datum].cols: continue
                    # For each column, get all other columns
                    interp:list[np.ndarray] = [interp1d(*data.get_xy_(daxisname, colname), # type: ignore
                                  kind="cubic",
                                  fill_value="extrapolate")(daxis) # type: ignore
                        for data in datas]
                    newcol = np.mean(interp, axis=0)
                    varcol = np.var(interp, axis=0)
                    stdcol = np.std(interp, axis=0)
                    outdata.add_col(colname, newcol, colkind, domain=datum)
                    outdata.add_col("v"+colname, varcol, colkind, domain=datum)
                    outdata.add_col("s"+colname, stdcol, colkind, domain=datum)
            newdata.add_data(outdata)
        
        context.data = newdata

        return CommandResult(True)

    @staticmethod
    def undo(args: NamedTuple, context: Context) -> CommandResult:
        # TODO:
        raise NotImplementedError



from __future__ import annotations

import numpy as np
import numpy.typing as npt

from dataclasses import dataclass
from logging import getLogger
from typing import NamedTuple
from scipy.interpolate import UnivariateSpline
from statsmodels.nonparametric.smoothers_lowess import lowess

from ._context import Context, Column, DataColType, Domain, range_to_index_
from ._handler import CommandHandler, Token, CommandResult
from ._numberunit import parse_nu, parse_range, actualize_range, NumberUnit, NumberUnitRange

from ._parser import CommandParser

from .fourier import fourier, Apodizer, get_flattop_window, bfourier

    

@dataclass(slots=True, frozen=True)
class Method:
    kweight: float

@dataclass(slots=True, frozen=True)
class Spline(Method):
    ...

@dataclass(slots=True, frozen=True)
class BSpline(Method):
    range: NumberUnitRange

@dataclass(slots=True, frozen=True)
class Fourier(Method):
    Rmax: float
    range: NumberUnitRange
    iter: int

@dataclass(slots=True, frozen=True)
class Smoothing(Method):
    range: NumberUnitRange
    fraction: float
    iterations: int

@dataclass(slots=True, frozen=True)
class Constant(Method):
    value: float

class Args_Background(NamedTuple):
    method:Method

def spline_method(x:npt.NDArray[np.floating], y:npt.NDArray[np.floating]) -> npt.NDArray[np.floating]:
    ...

def bspline_method(x:npt.NDArray[np.floating], y:npt.NDArray[np.floating]) -> npt.NDArray[np.floating]:
    s = UnivariateSpline(x,y,w=None,bbox=(None,None), k=3, s=None)
    return np.array(s(x))
    

class Background(CommandHandler):
    @staticmethod
    def parse(tokens: list[Token], context: Context) -> Args_Background:
        parser = CommandParser(
            "background",
            description="Calculates the free electron component of the signal.",
        )
        subparsers = parser.add_subparsers(dest="method")

        const = subparsers.add_parser("constant")
        const.add_argument("--value", "-v", type=float, default=1.0)

        spline = subparsers.add_parser("spline")
        spline.add_argument("--kweight", "-k", default=0, type=float)

        bspline = subparsers.add_parser("bspline")
        bspline.add_argument("range", nargs=2)
        bspline.add_argument("--kweight", "-k", default=0, type=float)

        fourier = subparsers.add_parser("fourier")
        fourier.add_argument("Rmax")
        fourier.add_argument("--range", nargs=2)
        fourier.add_argument("--kweight", "-k", default=0, type=float)
        fourier.add_argument("--iterations", "-i", type=int, default=3)

        smooth = subparsers.add_parser("smoothing")
        smooth.add_argument("range", nargs=2)
        smooth.add_argument("--kweight", "-k", default=0, type=float)
        smooth.add_argument("--fraction", "-f", type=float, default=0.3)
        smooth.add_argument("--iterations", "-i", type=int, default=1)


        args = parser.parse(tokens)

        match args.method:
            case "spline":
                method = Spline(args.kweight)
            case "constant":
                method = Constant(args.kweight, args.value)
            case "bspline":
                range = parse_range(*args.range)
                if range.domain != Domain.RECIPROCAL:
                    raise ValueError("Background must be fitted to the reciprocal domain.")
                method = BSpline(args.kweight, range)
            case "fourier":
                rmax = parse_nu(args.Rmax)
                if rmax.unit is None:
                    rmax = NumberUnit(rmax.value, rmax.sign, "A")
                if rmax.unit != "A": raise ValueError("The max distance must be in angstrom.")
                if args.range is not None:
                    range = parse_range(*args.range)
                    if range.domain != Domain.RECIPROCAL:
                        raise ValueError("Background must be fitted to the reciprocal domain.")
                else:
                    range = NumberUnitRange(NumberUnit(0, 0, "k"), NumberUnit(np.inf, 0, "k"), None, Domain.RECIPROCAL)
                method = Fourier(args.kweight, rmax.value, range, args.iterations)
            case "smoothing":
                range = parse_range(*args.range)
                if range.domain != Domain.RECIPROCAL:
                    raise ValueError("Background must be fitted to the reciprocal domain.")
                method = Smoothing(args.kweight, range, args.fraction, args.iterations)

        return Args_Background(method)

    @staticmethod
    def execute(args: Args_Background, context: Context) -> CommandResult:
        log = getLogger("background")

        match args.method:
            case Constant(kweight, value):
                for data in context.data:
                    X,Y = data.get_col_("k", domain=Domain.RECIPROCAL), data.get_col_("x", domain=Domain.RECIPROCAL)
                    S = np.full_like(X, value)
                    data.add_col("bkg", S, Column(None, None, DataColType.BACKGROUND), Domain.RECIPROCAL)
                    data.mod_col("x", Y-S)
                    log.debug(f"{data.meta.name}: Removed background as constant with value {value}.")
            
            case BSpline(kweight, _range):
                domain = _range.domain or Domain.RECIPROCAL
                if domain != Domain.RECIPROCAL:
                    raise RuntimeError("Cannot fit background to a different domain.")
        
                _axes = [data.get_col_("k") for data in context.data]
                bounds = actualize_range(_range, _axes, "k")

                for data in context.data:
                    X,Y = data.get_col_("k", domain=domain), data.get_col_("x", domain=domain)
                    idx = range_to_index_(data, bounds)
                    x, _y = X[idx], Y[idx]
                    y = _y * x ** kweight

                    s = bspline_method(x,y)
                    S = Y.copy()
                    S[idx] = s / x ** kweight

                    data.add_col("bkg", S, Column(None, None, DataColType.BACKGROUND), Domain.RECIPROCAL)
                    data.mod_col("x", Y-S)
                    log.debug(f"{data.meta.name}: Removed background as B-Spline contribution, with k-weight {kweight}.")

            
            case Fourier(kweight, Rmax, _range, iterations):
                domain = _range.domain or Domain.RECIPROCAL
                if domain != Domain.RECIPROCAL:
                    raise RuntimeError("Cannot fit background to a different domain.")
        
                _axes = [data.get_col_("k") for data in context.data]
                bounds = actualize_range(_range, _axes, "k")
                for data in context.data:
                    X,Y = data.get_col_("k", domain=Domain.RECIPROCAL), data.get_col_("x", domain=Domain.RECIPROCAL)
                    idx = range_to_index_(data, bounds)
                    x, _y = X[idx], Y[idx]
                    y = _y * x ** kweight
                    eps = 0.3
                    w = get_flattop_window(x, (x.max() - x.min())/8, Apodizer.HANN, None, 0, (-eps, +eps)) # type: ignore

                    r = np.linspace(-2*Rmax,2*Rmax, 1001)
                    W = get_flattop_window(r, Rmax/2, Apodizer.HANN, None, 0, (Rmax, -Rmax))

                    bkgs:list[npt.NDArray[np.floating]] = []
                    bkgs.append(np.real(bfourier(r, (fourier(x, (y)*w, r))*W, x))/2 / w) # type: ignore
                    
                    for _ in range(iterations):
                        bkgs.append(np.real(bfourier(r, (fourier(x, (y-sum(bkgs))*w, r))*W, x))/2 / w) # type: ignore
                    s = sum(bkgs)

                    S = Y.copy()
                    S[idx] = s / x ** kweight

                    data.add_col("bkg", S, Column(None, None, DataColType.BACKGROUND), Domain.RECIPROCAL)
                    data.mod_col("x", Y-S)
                    log.debug(f"{data.meta.name}: Removed background as Fourier contribution, up to {Rmax}A, with k-weight {kweight}.")
            
            case Smoothing(kweight, _range, fraction, iterations):
                domain = _range.domain or Domain.RECIPROCAL
                if domain != Domain.RECIPROCAL:
                    raise RuntimeError("Cannot fit postedge to a different domain.")
        
                _axes = [data.get_col_("k") for data in context.data]
                bounds = actualize_range(_range, _axes, "k")

                for data in context.data:
                    X,Y = data.get_col_("k", domain=domain), data.get_col_("x", domain=domain)
                    idx = range_to_index_(data, bounds)
                    x,_y = X[idx], Y[idx]
                    y = _y * x ** kweight

                    s = lowess(y, x, fraction, iterations, return_sorted=False)
                    S = Y.copy()
                    S[idx] = s / x ** kweight
                    data.add_col("bkg", S, Column(None, None, DataColType.BACKGROUND), Domain.RECIPROCAL)
                    data.mod_col("x", Y-S)
                    log.debug(f"{data.meta.name}: Removed background as smoothed contribution, with k-weight {kweight}.")
            case _:
                raise NotImplementedError("Method not implemented")

        return CommandResult(True)

    @staticmethod
    def undo(args: NamedTuple, context: Context) -> CommandResult:
        # TODO:
        raise NotImplementedError
